; *******************************************************
; *     CRT Interface Unit                              *
; *******************************************************

        TITLE   CRT

        INCLUDE SE.ASM

.8086

; Coordinate record
X       EQU     (BYTE PTR 0)
Y       EQU     (BYTE PTR 1)


DATA    SEGMENT WORD PUBLIC

; Externals
        extrn   ModeX:Byte,GraphMode:byte                 ; From SYSTEM
        EXTRN   CheckBreak:BYTE,CheckEOF:BYTE,DirectVideo:BYTE
        EXTRN   CheckSnow:BYTE,LastMode:WORD,ScanCode:byte
        EXTRN   WindMin:WORD,WindMax:WORD,Seg0040:WORD
        EXTRN   SegB000:WORD,SegB800:WORD
      
        extrn VideoBufferSave:dword
        extrn DelayCnt:dword, CRTLocal:byte ; Table of following:

include crt_save.inc

    IF DPMIVersion
        EXTRN   RealModeRegs:BYTE
    ENDIF

DATA    ENDS

CODE    SEGMENT BYTE PUBLIC

  ASSUME  CS:CODE,DS:DATA

; Externals
        EXTRN   HaltTurbo:FAR

; Publics
        PUBLIC  Initialize
      
        public CrtOldWrite
        public CrtOldReadKey,CrtOldKeyPressed
        public CrtSetup,AssignCrt
        public CrtInit,DelayLoop,GetCursor,SetCursor ; For CRT_2.ASM
      
        public CrtOldWindow,CrtOldGotoXY,CrtOldWhereX
        public CrtOldWhereY,CrtOldClrScr

; One-time initialization
Initialize:
        MOV     AH,0FH
        CALL    Video
        CMP     AL,7
              JE      @@1
        CMP     AL,3
        JBE     @@1
        MOV     AX,3
        CALL    CrtInit
@@1:    push cs
        CALL CrtSetup
        MOV     AH,8
        XOR     BH,BH
        CALL    Video
        MOV     AL,AH
        AND     AL,7FH
        MOV     NormAttr,AL
        MOV     TextAttr,AL
        XOR     AX,AX
        MOV     CheckEOF,AL
        MOV     BreakFlag,AL
        INC     AX
        MOV     CheckBreak,AL

;Next, calibrate Delay loop...
        MOV     ES,Seg0040
        MOV     DI,OFFSET Timer
        MOV     BL,ES:[DI]
@@2:    CMP     BL,ES:[DI]
        JE      @@2
@@3:    MOV     BL,ES:[DI]
        MOV     AX,-28
        CWD
        CALL    DelayLoop
        NOT     AX
        NOT     DX
        MOV     CX,55
        DIV     CX
        MOV     DelayCnt.w0,AX
      
          IF DPMIVersion
        MOV     AX,dpmiGetRMCB
        MOV     SI,OFFSET CtrlBreak
        MOV     DI,OFFSET RealModeRegs
              PUSH    DS
        POP     ES
        PUSH    CS
        POP     DS
        INT     DPMI
        PUSH    ES
        POP     DS
        MOV     AX,dpmiSetRealInt
        MOV     BL,1BH
        INT     DPMI
          ELSE
        PUSH    DS
         PUSH    CS
         POP     DS
         MOV     DX,OFFSET CtrlBreak
         MOV     AX,dosSetInt*256+1BH
         INT     DOS
        POP     DS
          ENDIF
        RETn

; Initialize CRT
; In    AL = Requested mode
;       AH = Requested font
CrtInit:mov ModeX, 0
        mov GraphMode, 0
        MOV     ES,Seg0040
        AND     ES:CrtInfo,0FEH
        CMP     AL,7
        JE      @@1
        CMP     AL,4
        JB      @@1
        MOV     AL,3
@@1:    PUSH    AX
        MOV     AH,0
        CALL    Video
        POP     AX
        OR      AH,AH
        JE      @@2
        MOV     AX,1112H
        MOV     BL,0
        CALL    Video
        MOV     AX,1130H
        MOV     BH,0
        MOV     DL,0
        CALL    Video
        CMP     DL,42
        JNE     @@2
        OR      ES:CrtInfo,1
        MOV     AX,100H
        MOV     CX,600H
        CALL    Video
        MOV     AH,12H
        MOV     BL,20H
        CALL    Video
@@2:    RET

; Setup CRT variables according to selected mode
CrtSetup:
        MOV     AH,0FH
        CALL    Video
        PUSH    AX
         MOV     AX,1130H
         MOV     BH,0
         MOV     DL,0
         CALL    Video
        POP     AX
        MOV     CL,0
        OR      DL,DL
        JNE     @@1
        MOV     DL,24
        CMP     AL,3
        JA      @@1
        MOV     CL,1
@@1:    MOV     DH,DL
        MOV     DL,AH
        DEC     DL
        MOV     AH,0
        CMP     DH,24
        JBE     @@2
        MOV     AH,1
@@2:    MOV     LastMode,AX
        MOV     CurCrtSize,DX
        MOV     CheckSnow,CL
        MOV     DirectVideo,1
        XOR     AX,AX
        MOV     WindMin,AX
        MOV     WindMax,DX
        RETf

; Ctrl-Break interrupt handler
CtrlBreak:
    IF DPMIVersion
        CMP     ES:CheckBreak,0
        JE      @@1
        MOV     ES:BreakFlag,1
@@1:    CLD
        LODSW
        MOV     ES:[DI].realIP,AX
        LODSW
        MOV     ES:[DI].realCS,AX
        LODSW
        MOV     ES:[DI].realFlags,AX
        ADD     ES:[DI].realSP,6
    ELSE
        PUSH    DS
         push SEG DATA
         pop ds
         CMP     CheckBreak,0
         JE      @@1
         MOV     BreakFlag,1
@@1:    POP     DS
    ENDIF
        IRET

; Check for Ctrl-Break
BreakCheck:
        CMP     BreakFlag,0
        JNE     @@1
        RETn
@@1:    MOV     BreakFlag,0
@@2:    MOV     AH,1
        INT     16H
        jz      @@3
        MOV     AH,0
        INT     16H
        JMP     @@2
@@3:    MOV     AL,'^'
        CALL    WriteChar
        MOV     AL,'C'
        CALL    WriteChar
        mov al, cr
        call WriteChar
        mov al, lf
        call WriteChar
        mov ax, 255
        jmp far ptr HaltTurbo

; Delay one timer tick or by CX iterations
DelayLoop:
@@1:    SUB     AX,1
        SBB     DX,0
        JC      @@2
        CMP     BL,ES:[DI]
        JE      @@1
@@2:    RETn

CrtOldKeyPressed:
        cmp ScanCode, 0
        jnz @@2
        mov ah, 11h
        int 16h
        mov al, 0
        jz @@1
@@2:    mov al, 1
@@1:    RETF

CrtOldReadKey:
        xor al, al
        xchg al, ScanCode
        test al, al
        jz @@1
        retf
@@1:    xor ax, ax
        int 16h
        test al, al
        jnz @@2
        mov ScanCode, ah
@@2:    call BreakCheck
        RETF

; Get cursor position
; Out   DX = Cursor position
; Uses  AX,BX,CX,DX,ES
GetCursor:
        MOV     AH,3
        XOR     BH,BH
        call near ptr Video
SaveCurs:mov byte ptr CursX, dl
        mov byte ptr CursY, dh
        xor bh, bh
        mov byte ptr CursX+1, bh
        mov byte ptr CursY+1, bh
        retn

; Set cursor position
; In    DX = Cursor position
; Uses  AX,BX,CX,DX,ES
SetCursor:
        MOV     AH,2
        ;XOR     BH,BH
        call near ptr SaveCurs
  ; Jatkuu Video()ssa

; Save registers and call BIOS video routine
Video:  PUSH    SI
         PUSH    DI
          PUSH    BP
           PUSH    ES
            INT     10H
           POP     ES
          POP     BP
         POP     DI
        POP     SI
        RETn

; Do pending direct write string
; In    BX    = Cursor position
;       ES:SI = String start address
;       ES:DI = String end address
; Uses  AX,BX,SI
DirectWrite:
        CMP     SI,DI
        JE      @@8
        PUSH    CX
         PUSH    DX
          PUSH    DI
           PUSH    DS
            PUSH    ES
             MOV     CX,DI
             SUB     CX,SI
             PUSH    DS
              mov ax, DATA
              mov ds, ax
              MOV     DS,Seg0040
              MOV     AL,BH
              MUL     DS:CrtWidth
              XOR     BH,BH
              ADD     AX,BX
              SHL     AX,1
              MOV     DI,AX
              MOV     DX,DS:Addr6845
              ADD     DX,6
              ;CMP     DS:CrtMode,7
             POP     DS
             MOV     AX,SegB800
             ;JNE     @@1
             ;MOV     AX,SegB000
@@1:         MOV     BL,CheckSnow
             MOV     BH,TextAttr
             PUSH    ES
             POP     DS
             MOV     ES,AX
             CLD
             OR      BL,BL
             JE      @@5
@@2:         LODSB
             MOV     BL,AL
@@3:         IN      AL,DX
             TEST    AL,1
             JNE     @@3
             CLI
@@4:         IN      AL,DX
             TEST    AL,1
             JE      @@4
             MOV     AX,BX
             STOSW
             STI
             LOOP    @@2
             JMP     SHORT @@7
@@5:         MOV     AH,BH
@@6:         LODSB
             STOSW
             LOOP    @@6
@@7:        POP     ES
           POP     DS
          POP     DI
         POP     DX
        POP     CX
@@8:    RETn

; Write character on CRT
; In    AL = Character
; Uses  AX
WriteChar:
        PUSH    BX
         PUSH    CX
          PUSH    DX
           PUSH    ES
            PUSH    AX
             CALL    GetCursor
            POP     AX
            CMP     AL,bell
            JE      @@1
            CMP     AL,bs
            JE      @@2
            CMP     AL,cr
            JE      @@3
            CMP     AL,lf
            JE      @@4
            cmp al, tab
            je @@6
            MOV     AH,9
            MOV     BL,TextAttr
            XOR     BH,BH
            MOV     CX,1
            PUSH    DX
             CALL    Video
            POP     DX
            INC     DL
            CMP     DL,WindMax.X
            JBE     @@5
            MOV     DL,WindMin.X
            JMP     SHORT @@4
@@1:        MOV     AH,14
            CALL    Video
            JMP     SHORT @@5
@@2:        CMP     DL,WindMin.X
            JE      @@5
            DEC     DL
            JMP     SHORT @@5
@@6:        mov al, ' '
            call WriteChar
            call GetCursor
            sub dl, WindMin.X
            and dl, 7
            jnz @@6
            call GetCursor
            jmp short @@5
@@3:        MOV     DL,WindMin.X
            JMP     SHORT @@5
@@4:        CALL    LineFeed
@@5:        CALL    SetCursor
           POP     ES
          POP     DX
         POP     CX
        POP     BX
        RETn

; Saves the uppermost line of window
; In   ds = seg data
; Uses ax,bx
SaveUppermost:
        mov ax, VideoBufferSave.w0
        mov bx, VideoBufferSave.w2
        push ax
         or ax, bx
        pop ax
        jz @@4                        ; Nil
        push es
         push di
          push si
           mov al, WindMin.Y
           mov es, Seg0040
           mul byte ptr es:CrtWidth
           add ax, ax
           les di, VideoBufferSave
           mov si, ax
           mov al, WindMin.X
           xor ah, ah
           add si, ax
      
           push cx
            mov cl, WindMax.X
            sub cl, WindMin.X
      
            push ds
             mov ds, SegB800

@@1:         jcxz @@2
             movsw
             jmp short @@1
@@2:        pop ds
           pop cx
          pop si
         pop di
         mov VideoBufferSave.w0, di
        pop es
@@4:    retn

; Do line-feed operation
; In    DX = Cursor position
; Uses  AX,BX
LineFeed:INC    DH
        CMP     DH,WindMax.Y
        JBE     @@5
        DEC     DH
      
        call SaveUppermost
      
        PUSH    CX
         PUSH    DX
          MOV     AX,601h
          MOV     BH,TextAttr
          MOV     CX,WindMin
          MOV     DX,WindMax
          CALL    Video
         POP     DX
        POP     CX
@@5:    RETn

; Write character string directly to CRT
; In    CX    = Character count
;       DX    = Position
;       ES:DI = String pointer
; Uses  AX,BX,CX,DX,SI,DI,ES
WriteString:
        PUSH    DS
         push ax
          mov ax, DATA
          mov ds, ax
         pop ax
         MOV     DS,Seg0040
         MOV     DX,DS:Cursor
        POP     DS
        MOV     BX,DX
        MOV     SI,DI
@@1:    MOV     AL,ES:[DI]
        CMP     AL,bell
        JE      @@2
        CMP     AL,bs
        JE      @@3
        CMP     AL,lf
        JE      @@4
        CMP     AL,cr
        JE      @@5
        cmp al, tab
        je @@9
        INC     DI
        INC     DL
        CMP     DL,WindMax.X
        JBE     @@8
        CALL    DirectWrite
        CALL    LineFeed
        MOV     DL,WindMin.X
        JMP     SHORT @@7
@@2:    CALL    DirectWrite
        PUSH    CX
         PUSH    DX
          MOV     AX,0E00h+bell
          CALL    Video
         POP     DX
        POP     CX
        JMP     SHORT @@6
@@3:    CALL    DirectWrite
        CMP     DL,WindMin.X
        JE      @@6
        DEC     DL
        JMP     SHORT @@6
@@4:    CALL    DirectWrite
        CALL    LineFeed
        JMP     SHORT @@6
@@9:    ; Tab ...
@@5:    CALL    DirectWrite
        MOV     DL,WindMin.X
@@6:    INC     DI
@@7:    MOV     SI,DI
        MOV     BX,DX
@@8:    LOOP    @@1
        CALL    DirectWrite
        PUSH    DS
         MOV     DS,Seg0040
         MOV     DS:Cursor,DX
         MOV     AL,DH
         MUL     DS:CrtWidth
         XOR     DH,DH
         ADD     AX,DX
         MOV     CX,AX
         MOV     DX,DS:Addr6845
         MOV     AL,14
         OUT     DX,AL
         JMP     SHORT $+2
         MOV     AL,CH
         INC     DX
         OUT     DX,AL
         JMP     SHORT $+2
         DEC     DX
         MOV     AL,15
         OUT     DX,AL
         JMP     SHORT $+2
         MOV     AL,CL
         INC     DX
         OUT     DX,AL
        POP     DS
        RETn

; Assign CRT to textfile
AssignCrt:
        MOV     BX,SP
        PUSH    DS
         LDS     DI,SS:[BX+4]
         MOV     [DI].fMode,fmClosed
         MOV     [DI].fBufSize,128
         LEA     AX,[DI].fBuffer
         MOV     [DI].fBufPtr.ofs,AX
         MOV     [DI].fBufPtr.seg,DS
         MOV     [DI].fOpenProc.ofs,OFFSET CrtOpen
         MOV     [DI].fOpenProc.seg,CS
         MOV     [DI].fName,0
        POP     DS
        RETF    4

; CRT file open procedure
CrtOpen:MOV     BX,SP
        PUSH    DS
         LDS     DI,SS:[BX+4]
         MOV     AX,OFFSET CrtRead
         MOV     BX,OFFSET CrtReturn
         MOV     [DI].fCloseProc.ofs,bx

         CMP     [DI].fMode,fmInput
         JE      @@1
         MOV     [DI].fMode,fmOutput
         MOV     AX,OFFSET CrtWrite
         MOV     BX,AX
@@1:     MOV     [DI].fInOutProc.ofs,AX
         MOV     [DI].fFlushProc.ofs,BX

         mov ax, cs
         MOV     [DI].fInOutProc.seg,ax
         MOV     [DI].fFlushProc.seg,ax
         MOV     [DI].fCloseProc.seg,ax
         XOR     AX,AX
        POP     DS
        RETF    4

; CRT file read procedure
CrtRead:ARG     FileP,DWORD,1
        ENTRY   FAR
        LES     DI,FileP
        MOV     DX,ES:[DI].fBufSize
        DEC     DX
        DEC     DX    ;DX=How many bytes the buffer may contain.
        MOV     SI,ES:[DI].fBufPos
        LES     DI,ES:[DI].fBufPtr
        XOR     BX,BX ;BX=How many bytes now in buffer.
@@1:    call @@10
        MOV     CX,1  ;Only one key currently.
        CMP     AL,bs
        JE      @@2
        cmp al, 'P'-64
        jne @@8
        push offset @@9
@@10:   push bx
         push cx
          push dx
           push si
            push di
             push es
              call dword ptr ReadKeyAddr
             pop es
            pop di
           pop si
          pop dx
         pop cx
        pop bx
        retn
@@8:    CMP     AL,'S'-64
        JE      @@2
        CMP     AL,'D'-64
        JE      @@3
        DEC     CX        ; Now the special keys alter the whole string.
        CMP     AL,esc
        JE      @@2
        CMP     AL,'A'-64 ; ^A = ESC...
        JE      @@2
        CMP     AL,'F'-64
        JE      @@3
        CMP     AL,eof
        JE      @@4
        CMP     AL,cr
        JE      @@5
        CMP     AL,' '
        JB      @@1
@@9:    CMP     BX,DX
        JE      @@1
        MOV     ES:[DI][BX],AL
        INC     BX            ;Now the buffer contains one char more.
        cmp LocalEcho, 1
        je @@13
        mov al, LocalEcho
@@13:   call @@7
        CMP     BX,SI
        JBE     @@1
        MOV     SI,BX
        JMP     @@1
@@2:    OR      BX,BX
        JZ      @@1
        mov al,bs
        call @@7
        mov al,' '
        call @@7
        mov al,bs
        call @@7
        DEC     BX
        LOOP    @@2
        JMP     @@1
@@3:    CMP     BX,SI
        JE      @@1
        MOV     AL,ES:[DI][BX]
        CMP     AL,' '
        JB      @@1
        call @@7
        INC     BX
        LOOP    @@3
        JMP     @@1
@@4:    CMP     CheckEOF,0
        jne @@11
        jmp @@1
@@11:   MOV     ES:[BX+DI],AL
        INC     BX
        JMP     SHORT @@6
@@7:    push bx
         push cx
          push es
           push di
            push si
             mov DummyChar, al
             cmp LocalEcho, 0
             jz @@12
             push 1
              push ds
               push offset DummyChar
             call dword ptr WriteStrProcAddr
@@12:       pop si
           pop di
          pop es
         pop cx
        pop bx
        retn
@@5:    push ax
         mov al, cr
         call @@7
         mov al, lf
         call @@7
        pop ax
        MOV     WORD PTR ES:[DI+BX],cr+lf*256
        INC     BX
        INC     BX
@@6:    LES     DI,FileP
        XOR     AX,AX
        MOV     ES:[DI].fBufPos,AX
        MOV     ES:[DI].fBufEnd,BX
        EXIT

; CRT file write procedure
CrtWrite:ARG FileP,DWORD,1
         ENTRY FAR
          LES DI,FileP
          xor cx, cx
          xchg cx, ES:[DI].fBufPos
          JCXZ An3
          LES DI,ES:[DI].fBufPtr
          push cx
           push es
            push di
          call dword ptr WriteStrProcAddr
An3:      call near ptr BreakCheck
          xor ax, ax
   EXIT

CrtOldWrite:
        arg StrSize,WORD, 1
        arg StrPtr,DWORD, 1
        entry far
         les di, StrPtr
         mov cx, StrSize
         jcxz An4
      
         CMP     DirectVideo,0
         JNE     An2
An1:     MOV     AL,ES:[DI]
         CALL    WriteChar
         INC     DI
         LOOP    An1
         jmp short An4
An2:     CALL    WriteString
An4:    exit

; CRT file no-op procedure
CrtReturn:
        XOR     AX,AX
        RETF    4

CrtOldGotoXY:
        MOV     BX,SP
        MOV     DL,SS:[BX+6]
        MOV     DH,SS:[BX+4]
        DEC     DL
        ADD     DL,WindMin.X
        JC      @@1
        CMP     DL,WindMax.X
        JA      @@1
        DEC     DH
        ADD     DH,WindMin.Y
        JC      @@1
        CMP     DH,WindMax.Y
        JA      @@1
        CALL    SetCursor
@@1:    RETF    4

CrtOldWhereX:
        ;mov ax, CursX
        CALL    GetCursor
        MOV     AL,DL
        xor ah, ah
        SUB     AL,WindMin.X
        sbb ah, 0
        INC     Ax
        RETF

CrtOldWhereY:
        ;mov ax, CursY
        CALL    GetCursor
        MOV     AL,DH
        xor ah, ah
        SUB     AL,WindMin.Y
        sbb ah, 0
        INC     Ax
        RETF

CrtOldClrScr:
        MOV     AX,600h
        MOV     BH,TextAttr
        MOV     CX,WindMin
        MOV     DX,WindMax
        CALL    Video
        MOV     DX,WindMin
        CALL    SetCursor
        RETF

CrtOldWindow:
        MOV     BX,SP
        MOV     DL,SS:[BX+10]
        MOV     DH,SS:[BX+8]
        MOV     CL,SS:[BX+6]
        MOV     CH,SS:[BX+4]
        CMP     DL,CL
        JA      @@1
        CMP     DH,CH
        JA      @@1
        DEC     DL
        JS      @@1
        DEC     DH
        JS      @@1
        DEC     CL
        CMP     CL,CurCrtSize.X
        JA      @@1
        DEC     CH
        CMP     CH,CurCrtSize.Y
        JA      @@1
        MOV     WindMin,DX
        MOV     WindMax,CX
        CALL    SetCursor
@@1:    RETF    8

CODE    ENDS

        END
