{*******************************************************}
{                                                       }
{       Windows DOS Interface Unit                      }
{                                                       }
{*******************************************************}
{ - lhes standardin mukainen. Lydt Turbo
 Pascalin helpeist kaiken tarvittavan. /BISQWIT}
unit WinDos;

{$IFNDEF Windows}
 {$IFDEF DPMI}
  {$L DOS_SWAP.OBP}           { Interrupt vector swapping }
 {$ELSE}
  {$L DOS_SWAP.OBJ}           { Interrupt vector swapping }
 {$ENDIF}
 {$L DOS_EXEC.OBJ}
{$ENDIF}

{$O+,S-,W-}
interface

{ String types }
Type
  ComStr  = string[127];        { Command line string }
  PathStr = string[79];         { File pathname string }
  DirStr  = string[67];         { Drive and directory string }
  NameStr = string[8];          { File name string }
  ExtStr  = string[4];          { File extension string }

{ Flags bit masks}
Const
  fCarry     = $0001;
  fParity    = $0004;
  fAuxiliary = $0010;
  fZero      = $0040;
  fSign      = $0080;
  fOverflow  = $0800;

{ File mode magic numbers}
Const
  fmClosed = $D7B0;
  fmInput  = $D7B1;
  fmOutput = $D7B2;
  fmInOut  = $D7B3;

{ File attribute Constants}
Const
  faReadOnly  = $01;
  faHidden    = $02;
  faSysFile   = $04;
  faVolumeID  = $08;
  faDirectory = $10;
  faArchive   = $20;
  faAnyFile   = $3F;

{ Maximum file name Cmponent string lengths}
Const
  fsPathName  = 79;
  fsDirectory = 67;
  fsFileName  = 8;
  fsExtension = 4;

{ FileSplit return flags}
Const
  fcExtension = $0001;
  fcFileName  = $0002;
  fcDirectory = $0004;
  fcWildcards = $0008;

{ Registers record used by Intr and MsDos}
type
  TRegisters = record
    case Integer of
      0: (AX, BX, CX, DX, BP, SI, DI, DS, ES, Flags: Word);
      1: (AL, AH, BL, BH, CL, CH, DL, DH: Byte);
  End;

{ Typed-file and untyped-file record}
Type
  TFileRec = Record
    Handle: Word;
    Mode: Word;
    RecSize: Word;
    Private: Array[1..26] of Byte;
    UserData: Array[1..16] of Byte;
    Name: Array[0..79] of Char;
  End;

{ Textfile record}
Type
  PTextBuf = ^TTextBuf;
  TTextBuf = Array[0..127]Of Char;
  TTextRec = Record
    Handle: Word;
    Mode: Word;
    BufSize: Word;
    Private: Word;
    BufPos: Word;
    BufEnd: Word;
    BufPtr: PTextBuf;
    OpenFunc: Pointer;
    InOutFunc: Pointer;
    FlushFunc: Pointer;
    CloseFunc: Pointer;
    UserData: Array[1..16]Of Byte;
    Name: Array[0..79]Of Char;
		Buffer: TTextBuf;
  End;

{ Search record used by FindFirst and FindNext}
type
  TSearchRec = record
    Fill: array[1..21] of Byte;
    Attr: Byte;
    Time: Longint;
    Size: Longint;
    Name: array[0..12] of Char;
  End;

{ Date and time record used by PackTime and UnpackTime}
type
  TDateTime = record
    Year, Month, Day, Hour, Min, Sec: Word;
  End;

{ Error status variable}
var
  DosError: Integer;

{ DosVersion returns the DOS version number. The low byte of    }
{ the result is the major version number, and the high byte is  }
{ the minor version number. For example, DOS 3.20 returns 3 in  }
{ the low byte, and 20 in the high byte.}
Function DosVersion: Word;

{ Intr executes a specified software interrupt with a specified }
{ TRegisters package. NOTE: To avoid general protection faults  }
{ when running in protected mode, always make sure to           }
{ initialize the DS and ES fields of the TRegisters record with }
{ valid selector values, or set the fields to zero.}
Procedure Intr(IntNo: Byte; var Regs: TRegisters);

{ MsDos invokes the DOS Function call handler with a specified  }
{ TRegisters package.}
Procedure MsDos(var Regs: TRegisters);

{ GetDate returns the current date set in the operating system. }
{ Ranges of the values returned are: Year 1980-2099, Month      }
{ 1-12, Day 1-31 and DayOfWeek 0-6 (0 corresponds to Sunday).}
Procedure GetDate(var Year, Month, Day, DayOfWeek: Word);

{ SetDate sets the current date in the operating system. Valid  }
{ parameter ranges are: Year 1980-2099, Month 1-12 and Day      }
{ 1-31. If the date is not valid, the Function call is ignored.}
Procedure SetDate(Year, Month, Day: Word);

{ GetTime returns the current time set in the operating system. }
{ Ranges of the values returned are: Hour 0-23, Minute 0-59,    }
{ Second 0-59 and Sec100 (hundredths of seconds) 0-99.}
Procedure GetTime(var Hour, Minute, Second, Sec100: Word);

{ SetTime sets the time in the operating system. Valid          }
{ parameter ranges are: Hour 0-23, Minute 0-59, Second 0-59 and }
{ Sec100 (hundredths of seconds) 0-99. If the time is not       }
{ valid, the Function call is ignored.}
Procedure SetTime(Hour, Minute, Second, Sec100: Word);

{ GetCBreak returns the state of Ctrl-Break checking in DOS.    }
{ When off (False), DOS only checks for Ctrl-Break during I/O   }
{ to console, printer, or communication devices. When on        }
{ (True), checks are made at every system call.}
Procedure GetCBreak(var Break: Boolean);

{ SetCBreak sets the state of Ctrl-Break checking in DOS.}
Procedure SetCBreak(Break: Boolean);

{ GetVerify returns the state of the verify flag in DOS. When   }
{ off (False), disk writes are not verified. When on (True),    }
{ all disk writes are verified to insure proper writing.}
Procedure GetVerify(var Verify: Boolean);

{ SetVerify sets the state of the verify flag in DOS.}
Procedure SetVerify(Verify: Boolean);

{ DiskFree returns the number of free bytes on the specified    }
{ drive number (0=Default,1=A,2=B,..). DiskFree returns -1 If   }
{ the drive number is invalid.}
Function DiskFree(Drive: Byte): Longint;

{ DiskSize returns the size in bytes of the specified drive     }
{ number (0=Default,1=A,2=B,..). DiskSize returns -1 If the     }
{ drive number is invalid.}
Function DiskSize(Drive: Byte): Longint;

{ GetFAttr returns the attributes of a file. F must be a file   }
{ variable (typed, untyped or textfile) which has been Assigned }
{ a name. The attributes are examined by ANDing with the        }
{ attribute masks defined as Constants above. Errors are        }
{ reported in DosError.}
Procedure GetFAttr(var F; var Attr: Word);

{ SetFAttr sets the attributes of a file. F must be a file      }
{ variable (typed, untyped or textfile) which has been Assigned }
{ a name. The attribute value is formed by adding (or ORing)    }
{ the appropriate attribute masks defined as Constants above.   }
{ Errors are reported in DosError.}
Procedure SetFAttr(var F; Attr: Word);

{ GetFTime returns the date and time a file was last written.   }
{ F must be a file variable (typed, untyped or textfile) which  }
{ has been Assigned and opened. The Time parameter may be       }
{ unpacked throgh a call to UnpackTime. Errors are reported in  }
{ DosError.}
Procedure GetFTime(var F; var Time: Longint);

{ SetFTime sets the date and time a file was last written.      }
{ F must be a file variable (typed, untyped or textfile) which  }
{ has been Assigned and opened. The Time parameter may be       }
{ created through a call to PackTime. Errors are reported in    }
{ DosError.}
Procedure SetFTime(var F; Time: Longint);

{ FindFirst searches the specified (or current) directory for   }
{ the first entry that matches the specified filename and       }
{ attributes. The result is returned in the specified search    }
{ record. Errors (and no files found) are reported in DosError.}
Function FindFirst(Path: PChar; Attr: Word; var F: TSearchRec) : Word;

{ FindNext returs the next entry that matches the name and      }
{ attributes specified in a previous call to FindFirst. The     }
{ search record must be one pAssed to FindFirst. Errors (and no }
{ more files) are reported in DosError.}
Function FindNext(var F: TSearchRec) : Word;

{ UnpackTime converts a 4-byte packed date/time returned by     }
{ FindFirst, FindNext or GetFTime into a TDateTime record.}
Procedure UnpackTime(P: Longint; var T: TDateTime);

{ PackTime converts a TDateTime record into a 4-byte packed     }
{ date/time used by SetFTime.}
Procedure PackTime(var T: TDateTime; var P: Longint);

{ GetIntVec returns the address stored in the specified         }
{ interrupt vector.}
Procedure GetIntVec(IntNo: Byte; var Vector: Pointer);

{ SetIntVec sets the address in the interrupt vector table for  }
{ the specified interrupt.}
Procedure SetIntVec(IntNo: Byte; Vector: Pointer);

{ FileSearch searches for the file given by Name in the list of }
{ directories given by List. The directory paths in List must   }
{ be separated by semicolons. The search always starts with the }
{ current directory of the current drive. If the file is found, }
{ FileSearch stores a concatenation of the directory path and   }
{ the file name in Dest. Otherwise FileSearch stores an empty   }
{ string in Dest. The maximum length of the result is defined   }
{ by the fsPathName Constant. The returned value is Dest.}
Function FileSearch(Dest, Name, List: PChar): PChar;

{ FileExpand fully expands the file name in Name, and stores    }
{ the result in Dest. The maximum length of the result is       }
{ defined by the fsPathName Constant. The result is an all      }
{ upper case string consisting of a drive letter, a colon, a    }
{ root relative directory path, and a file name. Embedded '.'   }
{ and '..' directory references are removed, and all name and   }
{ extension Cmponents are truncated to 8 and 3 characters. The }
{ returned value is Dest.}
Function FileExpand(Dest, Name: PChar): PChar;

{ FileSplit splits the file name specified by Path into its     }
{ three Cmponents. Dir is set to the drive and directory path  }
{ with any leading and trailing backslashes, Name is set to the }
{ file name, and Ext is set to the extension with a preceding   }
{ period. If a Cmponent string parameter is NIL, the           }
{ corresponding part of the path is not stored. If the path     }
{ does not contain a given Cmponent, the returned Cmponent    }
{ string is empty. The maximum lengths of the strings returned  }
{ in Dir, Name, and Ext are defined by the fsDirectory,         }
{ fsFileName, and fsExtension Constants. The returned value is  }
{ a combination of the fcDirectory, fcFileName, and fcExtension }
{ bit masks, indicating which Cmponents were present in the    }
{ path. If the name or extension contains any wildcard          }
{ characters (* or ?), the fcWildcards flag is set in the       }
{ returned value.}
Function FileSplit(Path, Dir, Name, Ext: PChar): Word;

{ GetCurDir returns the current directory of a specified drive. }
{ Drive = 0 indicates the current drive, 1 indicates drive A, 2 }
{ indicates drive B, and so on. The string returned in Dir      }
{ always starts with a drive letter, a colon, and a backslash.  }
{ The maximum length of the resulting string is defined by the  }
{ fsDirectory Constant. The returned value is Dir. Errors are   }
{ reported in DosError.}
Function GetCurDir(Dir: PChar; Drive: Byte): PChar;

{ SetCurDir changes the current directory to the path specified }
{ by Dir. If Dir specifies a drive letter, the current drive is }
{ also changed. Errors are reported in DosError.}
Procedure SetCurDir(Dir: PChar);

{ CreateDir creates a new subdirectory with the path specified  }
{ by Dir. Errors are reported in DosError.}
Procedure CreateDir(Dir: PChar);

{ RemoveDir removes the subdirectory with the path specified by }
{ Dir. Errors are reported in DosError.}
Procedure RemoveDir(Dir: PChar);

{ GetArgCount returns the number of parameters pAssed to the    }
{ program on the command line.}
Function GetArgCount: Integer;

{ GetArgStr returns the Index'th parameter from the command     }
{ line, or an empty string If Index is less than zero or        }
{ greater than GetArgCount. If Index is zero, GetArgStr returns }
{ the filename of the current module. The maximum length of the }
{ string returned in Dest is given by the MaxLen parameter. The }
{ returned value is Dest.}
Function GetArgStr(Dest: PChar; Index: Integer; MaxLen: Word): PChar;

{ GetEnvVar returns a pointer to the value of a specified       }
{ environment variable, i.e. a pointer to the first character   }
{ after the equals sign (=) in the environment entry given by   }
{ VarName. VarName is case insensitive. GetEnvVar returns NIL   }
{ If the specified environment variable does not exist.}
Function GetEnvVar(VarName: PChar): PChar;

{$IFNDEF Windows}
 Procedure Keep(Code: Word);
 procedure SwapVectors;
 procedure Exec(Path: PathStr; ComLine: ComStr);
 function DosExitCode: Word;
{$ENDIF}

implementation

{$IFDEF Windows}
{$DEFINE ProtectedMode}
{$ENDIF}
{$IFDEF DPMI}
{$DEFINE ProtectedMode}
{$ENDIF}
{$IFDEF Windows}
uses WinTypes, WinProcs, Strings;

{$ELSE}
uses Strings;

{$ENDIF}
{$IFDEF Windows}
Procedure AnsiDosFunc; Assembler;
var
  TempName: array[0..fsPathName] of Char;
Asm
        PUSH    DS
        PUSH    CX
        PUSH    AX
        MOV     SI,DI
        PUSH    ES
        POP     DS
        LEA     DI,TempName
        PUSH    SS
        POP     ES
        MOV     CX,fsPathName
        CLD
@@1:    LODSB
        OR      AL,AL
        JE      @@2
        STOSB
        LOOP    @@1
@@2:    XOR     AL,AL
        STOSB
        LEA     DI,TempName
        PUSH    SS
        PUSH    DI
        PUSH    SS
        PUSH    DI
        CALL    AnsiToOem
        POP     AX
        POP     CX
        LEA     DX,TempName
        PUSH    SS
        POP     DS
        INT     21H
        POP     DS
End;

{$ELSE}
Procedure AnsiDosFunc; Assembler;
Asm
        PUSH    DS
        MOV     DX,DI
        PUSH    ES
        POP     DS
        INT     21H
        POP     DS
End;

{$ENDIF}
Function DosVersion: Word; Assembler;
Asm
        MOV     AH,30H
        INT     21H
End;

Procedure Intr(IntNo: Byte; var Regs: TRegisters); Assembler;
Asm
        PUSH    DS
{$IFDEF ProtectedMode}
{$IFDEF Windows}
        PUSH    CS
        CALL    AllocCSToDSAlias
{$ELSE}
        MOV     AX,CS
        ADD     AX,SelectorInc
{$ENDIF}
        MOV     DS,AX
        CLI
        PUSH    WORD PTR DS:@@Int
        PUSH    DS
        MOV     AL,IntNo
        MOV     BYTE PTR DS:@@Int+1,AL
{$ELSE}
        PUSH    WORD PTR CS:@@Int
        MOV     AL,IntNo
        MOV     BYTE PTR CS:@@Int+1,AL
{$ENDIF}
        LDS     SI,Regs
        CLD
        LODSW
        PUSH    AX
        LODSW
        XCHG    AX,BX
        LODSW
        XCHG    AX,CX
        LODSW
        XCHG    AX,DX
        LODSW
        XCHG    AX,BP
        LODSW
        PUSH    AX
        LODSW
        XCHG    AX,DI
        LODSW
        PUSH    AX
        LODSW
{$IFDEF DPMI}
        VERR    AX
        JNZ     @@1
        MOV     ES,AX
@@1:    POP     AX
        VERR    AX
        JNZ     @@2
        MOV     DS,AX
@@2:
{$ELSE}
        MOV     ES,AX
        POP     DS
{$ENDIF}
        POP     SI
        POP     AX
@@Int:  INT     0
        STI
        PUSHF
        PUSH    ES
        PUSH    DI
        PUSH    BP
        MOV     BP,SP
{$IFDEF ProtectedMode}
        LES     DI,Regs+14
{$ELSE}
        LES     DI,Regs+12
{$ENDIF}
        CLD
        STOSW
        XCHG    AX,BX
        STOSW
        XCHG    AX,CX
        STOSW
        XCHG    AX,DX
        STOSW
        POP     AX
        STOSW
        XCHG    AX,SI
        STOSW
        POP     AX
        STOSW
        MOV     AX,DS
        STOSW
        POP     AX
        STOSW
        POP     AX
        STOSW
{$IFDEF ProtectedMode}
        POP     DS
        POP     WORD PTR DS:@@Int
{$ELSE}
        POP     WORD PTR CS:@@Int
{$ENDIF}
{$IFDEF Windows}
        MOV     AX,DS
        POP     DS
        PUSH    AX
        CALL    FreeSelector
{$ELSE}
        POP     DS
{$ENDIF}
End;

Procedure MsDos(var Regs: TRegisters);
begin
  Intr($21, Regs);
End;

Procedure GetDate(var Year, Month, Day, DayOfWeek: Word); Assembler;
Asm
        MOV     AH,2Ah
        INT     21H
        XOR     AH,AH
        LES     DI,DayOfWeek
        STOSW
        MOV     AL,DL
        LES     DI,Day
        STOSW
        MOV     AL,DH
        LES     DI,Month
        STOSW
        XCHG    AX,CX
        LES     DI,Year
        STOSW
End;

Procedure SetDate(Year, Month, Day: Word); Assembler;
Asm
        MOV     CX,Year
        MOV     DH,BYTE PTR Month
        MOV     DL,BYTE PTR Day
        MOV     AH,2Bh
        INT     21H
End;

Procedure GetTime(var Hour, Minute, Second, Sec100: Word); Assembler;
Asm
        MOV     AH,2CH
        INT     21H
        XOR     AH,AH
        MOV     AL,DL
        LES     DI,Sec100
        STOSW
        MOV     AL,DH
        LES     DI,Second
        STOSW
        MOV     AL,CL
        LES     DI,Minute
        STOSW
        MOV     AL,CH
        LES     DI,Hour
        STOSW
End;

Procedure SetTime(Hour, Minute, Second, Sec100: Word); Assembler;
Asm
        MOV     CH,BYTE PTR Hour
        MOV     CL,BYTE PTR Minute
        MOV     DH,BYTE PTR Second
        MOV     DL,BYTE PTR Sec100
        MOV     AH,2DH
        INT     21H
End;

Procedure GetCBreak(var Break: Boolean); Assembler;
Asm
        MOV     AX,3300h
        INT     21H
        LES     DI,Break
        MOV     ES:[DI],DL
End;

Procedure SetCBreak(Break: Boolean); Assembler;
Asm
        MOV     DL,Break
        MOV     AX,3301h
        INT     21H
End;

Procedure GetVerify(var Verify: Boolean); Assembler;
Asm
        MOV     AH,54h
        INT     21H
        LES     DI,Verify
        STOSB
End;

Procedure SetVerify(Verify: Boolean); Assembler;
Asm
        MOV     AL,Verify
        MOV     AH,2Eh
        INT     21H
End;

Function DiskFree(Drive: Byte): Longint; Assembler;
Asm
        MOV     DL,Drive
        MOV     AH,36h
        INT     21H
        MOV     DX,AX
        CMP     AX,0FFFFH
        JE      @@1
        MUL     CX
        MUL     BX
@@1:
End;

Function DiskSize(Drive: Byte): Longint; Assembler;
Asm
        MOV     DL,Drive
        MOV     AH,36h
        INT     21H
        MOV     BX,DX
        MOV     DX,AX
        CMP     AX,0FFFFH
        JE      @@1
        MUL     CX
        MUL     BX
@@1:
End;

Procedure GetFAttr(var F; var Attr: Word); Assembler;
Asm
        PUSH    DS
         LDS     DX,F
         ADD     DX,OFFSET TFileRec.Name
         MOV     AX,4300h
         INT     21H
        POP     DS
        JNC     @@1
        XOR     CX,CX
        JMP     @@2
@@1:    XOR     AX,AX
@@2:    MOV     DosError,AX
        LES     DI,Attr
        XCHG    AX,CX
        STOSW
End;

Procedure SetFAttr(var F; Attr: Word); Assembler;
Asm
        PUSH    DS
         LDS     DX,F
         ADD     DX,OFFSET TFileRec.Name
         MOV     CX,Attr
         MOV     AX,4301H
         INT     21H
        POP     DS
        JC      @@1
        XOR     AX,AX
@@1:    MOV     DosError,AX
End;

Procedure GetFTime(var F; var Time: Longint); Assembler;
Asm
        LES     DI,F
        MOV     BX,ES:[DI].TFileRec.Handle
        MOV     AX,5700H
        INT     21H
        JNC     @@1
        XOR     CX,CX
        XOR     DX,DX
        JMP     @@2
@@1:    XOR     AX,AX
@@2:    MOV     DosError,AX
        LES     DI,Time
        CLD
        XCHG    AX,CX
        STOSW
        XCHG    AX,DX
        STOSW
End;

Procedure SetFTime(var F; Time: Longint); Assembler;
Asm
        LES     DI,F
        MOV     BX,ES:[DI].TFileRec.Handle
        MOV     CX,WORD PTR Time[0]
        MOV     DX,WORD PTR Time[2]
        MOV     AX,5701H
        INT     21H
        JC      @@1
        XOR     AX,AX
@@1:    MOV     DosError,AX
End;

Function FindFirst; Assembler; {Bisqwit}
Asm
				PUSH    DS
				LDS     DX,F
				MOV     AH,1AH
				INT     21H
				POP     DS
				LES     DI,Path
				MOV     CX,Attr
				MOV     AH,4EH
				CALL    AnsiDosFunc
				JC      @@1
{$IFDEF Windows}
				LES     DI,F
				ADD     DI,OFFSET TSearchRec.Name
				PUSH    ES
				PUSH    DI
				PUSH    ES
				PUSH    DI
				CALL    OemToAnsi
{$ENDIF}
				XOR     AX,AX
@@1:    MOV     DosError,AX
End;

Function FindNext; Assembler; {Bisqwit}
Asm
        PUSH    DS
        LDS     DX,F
        MOV     AH,1AH
        INT     21H
        POP     DS
        MOV     AH,4FH
        INT     21H
        JC      @@1
{$IFDEF Windows}
        LES     DI,F
        ADD     DI,OFFSET TSearchRec.Name
        PUSH    ES
        PUSH    DI
        PUSH    ES
        PUSH    DI
        CALL    OemToAnsi
{$ENDIF}
				XOR     AX,AX
@@1:    MOV     DosError,AX
End;

Procedure UnpackTime(P: Longint; var T: TDateTime); Assembler;
Asm
        LES     DI,T
        CLD
        MOV     AX,P.Word[2]
        MOV     CL,9
        SHR     AX,CL
        ADD     AX,1980
        STOSW
        MOV     AX,P.Word[2]
        MOV     CL,5
        SHR     AX,CL
        AND     AX,15
        STOSW
        MOV     AX,P.Word[2]
        AND     AX,31
        STOSW
        MOV     AX,P.Word[0]
        MOV     CL,11
        SHR     AX,CL
        STOSW
        MOV     AX,P.Word[0]
        MOV     CL,5
        SHR     AX,CL
        AND     AX,63
        STOSW
        MOV     AX,P.Word[0]
        AND     AX,31
        SHL     AX,1
        STOSW
End;

Procedure PackTime(var T: TDateTime; var P: Longint); Assembler;
Asm
        PUSH    DS
         LDS     SI,T
         CLD
         LODSW
         SUB     AX,1980
         MOV     CL,9
         SHL     AX,CL
         XCHG    AX,DX
         LODSW
         MOV     CL,5
         SHL     AX,CL
         ADD     DX,AX
         LODSW
         ADD     DX,AX
         LODSW
         MOV     CL,11
         SHL     AX,CL
         XCHG    AX,BX
         LODSW
         MOV     CL,5
         SHL     AX,CL
         ADD     BX,AX
         LODSW
         SHR     AX,1
         ADD     AX,BX
        POP     DS
        LES     DI,P
        STOSW
        XCHG    AX,DX
        STOSW
End;

Procedure GetIntVec(IntNo: Byte; var Vector: Pointer); Assembler;
Asm
        MOV     AL,IntNo
        MOV     AH,35h
        INT     21H
        MOV     AX,ES
        LES     DI,Vector
        CLD
        XCHG    AX,BX
        STOSW
        XCHG    AX,BX
        STOSW
End;

Procedure SetIntVec(IntNo: Byte; Vector: Pointer); Assembler;
Asm
        PUSH    DS
         LDS     DX,Vector
         MOV     AL,IntNo
         MOV     AH,25h
         INT     21H
        POP     DS
End;

Function FileSearch(Dest, Name, List: PChar): PChar; Assembler;
Asm
        PUSH    DS
         CLD
         LDS     SI,List
         LES     DI,Dest
         MOV     CX,fsPathName
@@1:     PUSH    DS
          PUSH    SI
           JCXZ    @@3
           LDS     SI,Name
@@2:       LODSB
           OR      AL,AL
           JE      @@3
           STOSB
           LOOP    @@2
@@3:       XOR     AL,AL
           STOSB
           LES     DI,Dest
           MOV     AX,4300H
           CALL    AnsiDosFunc
          POP     SI
         POP     DS
         JC      @@4
         TEST    CX,18H
         JE      @@9
@@4:     LES     DI,Dest
         MOV     CX,fsPathName
         XOR     AH,AH
         LODSB
         OR      AL,AL
         JE      @@8
@@5:     CMP     AL,';'
         JE      @@7
         JCXZ    @@6
         MOV     AH,AL
         STOSB
         DEC     CX
@@6:     LODSB
         OR      AL,AL
         JNE     @@5
         DEC     SI
@@7:     JCXZ    @@1
         CMP     AH,':'
         JE      @@1
         MOV     AL,'\'
         CMP     AL,AH
         JE      @@1
         STOSB
         DEC     CX
         JMP     @@1
@@8:     STOSB
@@9:     MOV     AX,Dest.Word[0]
         MOV     DX,Dest.Word[2]
        POP     DS
End;

Function FileExpand(Dest, Name: PChar): PChar; Assembler;
(*var
  TempName: array[0..159] of Char;*)
Asm
        PUSH    DS
        mov ah, 60h
        lds si, Name
        les di, Dest
        int 21h
        mov ax, di
        mov dx, es
(*        CLD
        LDS     SI,Name
        LEA     DI,TempName
        PUSH    SS
        POP     ES
        LODSW
        OR      AL,AL
        JE      @@1
        CMP     AH,':'
        JNE     @@1
        CMP     AL,'a'
        JB      @@2
        CMP     AL,'z'
        JA      @@2
        SUB     AL,20H
        JMP     @@2
@@1:    DEC     SI
        DEC     SI
        MOV     AH,19H
        INT     21H
        ADD     AL,'A'
        MOV     AH,':'
@@2:    STOSW
        CMP     [SI].Byte,'\'
        JE      @@3
        SUB     AL,'A'-1
        MOV     DL,AL
        MOV     AL,'\'
        STOSB
        PUSH    DS
        PUSH    SI
        MOV     AH,47H
        MOV     SI,DI
        PUSH    ES
        POP     DS
        INT     21H
        POP     SI
        POP     DS
        JC      @@3
        XOR     AL,AL
        CMP     AL,ES:[DI]
        JE      @@3
{$IFDEF Windows}
        PUSH    ES
        PUSH    ES
        PUSH    DI
        PUSH    ES
        PUSH    DI
        CALL    OemToAnsi
        POP     ES
{$ENDIF}
        MOV     CX,0FFFFH
        XOR     AL,AL
        CLD
        REPNE   SCASB
        DEC     DI
        MOV     AL,'\'
        STOSB
@@3:    MOV     CX,8
@@4:    LODSB
        OR      AL,AL
        JE      @@7
        CMP     AL,'\'
        JE      @@7
        CMP     AL,'.'
        JE      @@6
        JCXZ    @@4
        DEC     CX
{$IFNDEF Windows}
        CMP     AL,'a'
        JB      @@5
        CMP     AL,'z'
        JA      @@5
        SUB     AL,20H
{$ENDIF}
@@5:    STOSB
        JMP     @@4
@@6:    MOV     CL,3
        JMP     @@5
@@7:    CMP     ES:[DI-2].Word,'.\'
        JNE     @@8
        DEC     DI
        DEC     DI
        JMP     @@10
@@8:    CMP     ES:[DI-2].Word,'..'
        JNE     @@10
        CMP     ES:[DI-3].Byte,'\'
        JNE     @@10
        SUB     DI,3
        CMP     ES:[DI-1].Byte,':'
        JE      @@10
@@9:    DEC     DI
        CMP     ES:[DI].Byte,'\'
        JNE     @@9
@@10:   MOV     CL,8
        OR      AL,AL
        JNE     @@5
        CMP     ES:[DI-1].Byte,':'
        JNE     @@11
        MOV     AL,'\'
        STOSB
@@11:   LEA     SI,TempName
        PUSH    SS
        POP     DS
        MOV     CX,DI
        SUB     CX,SI
        CMP     CX,79
        JBE     @@12
        MOV     CX,79
@@12:   LES     DI,Dest
        PUSH    ES
        PUSH    DI
{$IFDEF Windows}
        PUSH    ES
        PUSH    DI
{$ENDIF}
        REP     MOVSB
        XOR     AL,AL
        STOSB
{$IFDEF Windows}
        CALL    AnsiUpper
{$ENDIF}
        POP     AX
        POP     DX*)
        POP     DS
End;

{$W+}
Function FileSplit(Path, Dir, Name, Ext: PChar): Word;
var
  DirLen, NameLen, Flags: Word;
  NamePtr, ExtPtr: PChar;
begin
  NamePtr := StrRScan(Path, '\');
  If NamePtr = nil then NamePtr := StrRScan(Path, ':');
  If NamePtr = nil then NamePtr := Path else Inc(NamePtr);
  ExtPtr := StrScan(NamePtr, '.');
  If ExtPtr = nil then ExtPtr := StrEnd(NamePtr);
  DirLen := NamePtr - Path;
  If DirLen > fsDirectory then DirLen := fsDirectory;
  NameLen := ExtPtr - NamePtr;
  If NameLen > fsFilename then NameLen := fsFilename;
  Flags := 0;
  If(StrScan(NamePtr, '?') <> nil) or (StrScan(NamePtr, '*') <> nil) then
    Flags := fcWildcards;
  If DirLen <> 0 then Flags := Flags or fcDirectory;
  If NameLen <> 0 then Flags := Flags or fcFilename;
  If ExtPtr[0] <> #0 then Flags := Flags or fcExtension;
  If Dir <> nil then StrNCpy(Dir, Path, DirLen);
  If Name <> nil then StrNCpy(Name, NamePtr, NameLen);
  If Ext <> nil then StrNCpy(Ext, ExtPtr, fsExtension);
  FileSplit := Flags;
End;

{$W-}
Function GetCurDir(Dir: PChar; Drive: Byte): PChar; Assembler;
Asm
        MOV     AL,Drive
        OR      AL,AL
        JNE     @@1
        MOV     AH,19H
        INT     21H
        INC     AX
@@1:    MOV     DL,AL
        LES     DI,Dir
        PUSH    ES
        PUSH    DI
        CLD
        ADD     AL,'A'-1
        MOV     AH,':'
        STOSW
        MOV     AX,'\'
        STOSW
        PUSH    DS
        LEA     SI,[DI-1]
        PUSH    ES
        POP     DS
        MOV     AH,47H
        INT     21H
        JC      @@2
{$IFDEF Windows}
        PUSH    DS
        PUSH    SI
        PUSH    DS
        PUSH    SI
        CALL    OemToAnsi
{$ENDIF}
        XOR     AX,AX
@@2:    POP     DS
        MOV     DosError,AX
        POP     AX
        POP     DX
End;

Procedure SetCurDir(Dir: PChar); Assembler;
Asm
        LES     DI,Dir
        cmp SetDriveAndPath, 0
        jz @@1
        MOV     AX,ES:[DI]
        OR      AL,AL
        JE      @@2
        CMP     AH,':'
        JNE     @@1
        AND     AL,0DFH
        SUB     AL,'A'
        MOV     DL,AL
        MOV     AH,0EH
        INT     21H
        MOV     AH,19H
        INT     21H
        CMP     AL,DL
        MOV     AX,15
        JNE     @@3
        CMP     AH,ES:[DI+2]
        JE      @@2
@@1:    MOV     AH,3BH
        CALL    AnsiDosFunc
        JC      @@3
@@2:    XOR     AX,AX
@@3:    MOV     DosError,AX
End;

Procedure CreateDir(Dir: PChar); Assembler;
Asm
        LES     DI,Dir
        MOV     AH,39H
        CALL    AnsiDosFunc
        JC      @@1
        XOR     AX,AX
@@1:    MOV     DosError,AX
End;

Procedure RemoveDir(Dir: PChar); Assembler;
Asm
        LES     DI,Dir
        MOV     AH,3AH
        CALL    AnsiDosFunc
        JC      @@1
        XOR     AX,AX
@@1:    MOV     DosError,AX
End;

{$IFDEF Windows}
Procedure ArgStrCount; Assembler;
Asm
        LDS     SI,CmdLine
        CLD
@@1:    LODSB
        OR      AL,AL
        JE      @@2
        CMP     AL,' '
        JBE     @@1
@@2:    DEC     SI
        MOV     BX,SI
@@3:    LODSB
        CMP     AL,' '
        JA      @@3
        DEC     SI
        MOV     AX,SI
        SUB     AX,BX
        JE      @@4
        LOOP    @@1
@@4:
End;

Function GetArgCount: Integer; Assembler;
Asm
        PUSH    DS
         XOR     CX,CX
         CALL    ArgStrCount
         XCHG    AX,CX
         NEG     AX
        POP     DS
End;

Function GetArgStr(Dest: PChar; Index: Integer;
        MaxLen: Word): PChar; Assembler;
Asm
        MOV     CX,Index
        JCXZ    @@2
        PUSH    DS
        CALL    ArgStrCount
        MOV     SI,BX
        LES     DI,Dest
        MOV     CX,MaxLen
        CMP     CX,AX
        JB      @@1
        XCHG    AX,CX
@@1:    REP     MOVSB
        XCHG    AX,CX
        STOSB
        POP     DS
        JMP     @@3
@@2:    PUSH    HInstance
        PUSH    Dest.Word[2]
        PUSH    Dest.Word[0]
        MOV     AX,MaxLen
        INC     AX
        PUSH    AX
        CALL    GetModuleFileName
@@3:    MOV     AX,Dest.Word[0]
        MOV     DX,Dest.Word[2]
End;

{$ELSE}
Procedure ArgStrCount; Assembler;
Asm
        MOV     DS,PrefixSeg
        MOV     SI,80H
        CLD
        LODSB
        MOV     DL,AL
        XOR     DH,DH
        ADD     DX,SI
@@1:    CMP     SI,DX
        JE      @@2
        LODSB
        CMP     AL,' '
        JBE     @@1
        DEC     SI
@@2:    MOV     BX,SI
@@3:    CMP     SI,DX
        JE      @@4
        LODSB
        CMP     AL,' '
        JA      @@3
        DEC     SI
@@4:    MOV     AX,SI
        SUB     AX,BX
        JE      @@5
        LOOP    @@1
@@5:
End;

Function GetArgCount: Integer; Assembler;
Asm
        PUSH    DS
         XOR     CX,CX
         CALL    ArgStrCount
         XCHG    AX,CX
         NEG     AX
        POP     DS
End;

Function GetArgStr(Dest:PChar; Index:Integer; MaxLen:Word): PChar; Assembler;
Asm
        PUSH    DS
         MOV     CX,Index
         JCXZ    @@1
         CALL    ArgStrCount
         MOV     SI,BX
         JMP     @@4
@@1:     MOV     AH,30H
         INT     21H
         CMP     AL,3
         MOV     AX,0
         JB      @@4
         MOV     DS,PrefixSeg
         MOV     ES,DS:WORD PTR 2CH
         XOR     DI,DI
         CLD
@@2:     CMP     AL,ES:[DI]
         JE      @@3
         MOV     CX,-1
         REPNE   SCASB
         JMP     @@2
@@3:     ADD     DI,3
         MOV     SI,DI
         PUSH    ES
         POP     DS
         MOV     CX,256
         REPNE   SCASB
         XCHG    AX,CX
         NOT     AL
@@4:     LES     DI,Dest
         MOV     CX,MaxLen
         CMP     CX,AX
         JB      @@5
         XCHG    AX,CX
@@5:     REP     MOVSB
         XCHG    AX,CX
         STOSB
         MOV     AX,Dest.Word[0]
         MOV     DX,Dest.Word[2]
        POP     DS
End;

{$ENDIF}
{$W+}
Function GetEnvVar(VarName: PChar): PChar;
var
  L: Word;
  P: PChar;
begin
  L := StrLen(VarName);
{$IFDEF Windows}
  P := GetDosEnvironment;
{$ELSE}
  P := Ptr(Word(Ptr(PrefixSeg, $2C)^), 0);
{$ENDIF}
  while P^ <> #0 do
  begin
    If(StrNICmp(P, VarName, L) = 0) and (P[L] = '=') then
    begin
      GetEnvVar := P + L + 1;
      Exit;
    End;
    Inc(P, StrLen(P) + 1);
  End;
  GetEnvVar := nil;
End;

{$W-}

{$IFNDEF Windows}
procedure SwapVectors; external {SWAP};
procedure Exec; external {EXEC};
Function DosExitCode; external {EXEC};

Procedure Keep(Code: Word); Assembler;
Asm
  mov ax, PrefixSeg
  mov es, ax
  mov dx, word ptr es:[2]
  sub dx, ax
  mov ax, Code
  mov ah, 31h
  int 21h
End;
{$ENDIF}
End.
